"""Client for basic usage"""

import logging
import os
from typing import AsyncGenerator, Optional, Type, TypeVar, Union, cast

from google.cloud.firestore import AsyncClient
from google.oauth2.service_account import Credentials
from google.api_core.exceptions import AlreadyExists

from frantic.base import BaseModel
from frantic import exceptions


ModelType = TypeVar("ModelType", bound=BaseModel)  # pylint: disable=invalid-name

logger = logging.getLogger(__name__)


def client_factory() -> AsyncClient:
    """Create a default client instance"""
    gappcreds = os.getenv("GOOGLE_APPLICATION_CREDENTIALS")
    if gappcreds is None:
        raise ValueError("No GOOGLE_APPLICATION_CREDENTIALS env was set")
    credentials = Credentials.from_service_account_file(gappcreds)
    return AsyncClient(credentials=credentials)


class Frantic:
    """Firestore with Pydantic models integration"""

    def __init__(self, *, client: Optional[AsyncClient] = None, prefix: Optional[str] = None):
        self.client = client or client_factory()
        self.prefix = (prefix or "").rstrip("/").lstrip("/")

    def _get_path(self, model: Type[ModelType]) -> str:
        """Get collection path for a specified model"""
        colname = model.collection or model.__name__.lower()
        return f"{self.prefix}/{colname}" if self.prefix else colname

    async def get(self, model: Type[ModelType], did: str) -> Optional[ModelType]:
        """Get a single document of a type {model} by its document id {did}

        Returns None if none found
        """
        collection_path = self._get_path(model)
        collection = self.client.collection(collection_path)
        docref = collection.document(document_id=did)
        document = await docref.get()
        if not document.exists:
            return None
        return model(**document.to_dict(), id=document.id)

    # TODO: Pagination
    async def list(self, model: Type[ModelType]) -> AsyncGenerator[ModelType, None]:
        """Iterate through all documents of a type {model}"""
        collection_path = self._get_path(model)
        collection = self.client.collection(collection_path)
        async for docref in collection.list_documents():
            document = await docref.get()
            yield model(**document.to_dict(), id=docref.id)

    async def add(self, instance: ModelType) -> ModelType:
        """Add instance as a document within corresponding collection"""
        collection_path = self._get_path(instance.__class__)
        collection = self.client.collection(collection_path)
        try:
            _, docref = await collection.add(
                instance.dict(exclude={"id"}), document_id=instance.id if instance.id is not None else None
            )
        except AlreadyExists as error:
            raise exceptions.AlreadyExists(instance.id) from error
        instance.id = docref.id
        return instance

    async def delete(self, model_or_instance: Union[Type[ModelType], ModelType], did: Optional[str] = None):
        """Delete a document specified by model type and document id {did}"""
        if isinstance(model_or_instance, BaseModel):
            if did:
                logger.warning(
                    "Document ID was provided even though an instance was provided as well. Using ID of the instance."
                )
            did = model_or_instance.id
            model = type(model_or_instance)
        else:
            model = cast(Type[BaseModel], model_or_instance)
        if did is None:
            raise ValueError(f"Cannot delete '{model.__name__}' without an id")
        collection_path = self._get_path(model)
        collection = self.client.collection(collection_path)
        docref = collection.document(did)
        await docref.delete()



"""Pydantic base with Frantic capabilities
"""

from typing import ClassVar, Optional

from pydantic import BaseModel as _BaseModel  # pylint: disable=no-name-in-module


# pylint: disable=too-few-public-methods
class BaseModel(_BaseModel):
    """Frantic base model"""

    collection: ClassVar[Optional[str]] = None
    id: Optional[str] = None



"""Version from pyproject.toml
"""

import os
import tomli

with open(os.path.join(os.path.dirname(__file__), "../pyproject.toml"), "rb") as fid:
    pyproject = tomli.load(fid)

__version__ = pyproject["tool"]["poetry"]["version"]

__all__ = ["__version__"]



"""Exceptions"""


class FranticException(Exception):
    """A base for all Frantic Exceptions"""


class AlreadyExists(FranticException):
    """Raised on document conflict"""

    def __init__(self, document_id: str | None):
        self.document_id = document_id
        super().__init__(f"Document with id {self.document_id} already exists")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['frantic']

package_data = \
{'': ['*']}

install_requires = \
['google-cloud-firestore>=2.6.0,<3.0.0', 'pydantic>=1.9.1,<2.0.0']

setup_kwargs = {
    'name': 'frantic',
    'version': '0.1.3',
    'description': 'Firestore with Pydantic models integration',
    'long_description': '# Frantic\n\nFirestore with Pydantic models integration.\n\nPlease note that this module is a work in progress. API may change over time. Also, the code is not tested yet\nand there are no security checks run periodically.\n\n## Installation\n\n### Using `pip`\n\n```console\npip install frantic\n```\n\n### Using `poetry`\n\n```console\npoetry add frantic\n```\n\n## Basic usage\n\nIf you have your service account key path set in `GOOGLE_APPLICATION_CREDENTIALS`, it\'s fairly easy:\n\n```python\nimport asyncio\nfrom typing import ClassVar\nfrom frantic import Frantic, BaseModel\n\n\n# Create a model the same way you would do it with Pydantic\nclass User(BaseModel):\n    # optionally, you may specify name of collection instances will be stored within:\n    collection: ClassVar = "my_users_collection"\n    # field \'id\' is added automatically\n    name: str\n\n\nasync def main():\n    frantic = Frantic()\n\n    user = User(name="ijustfarted")\n\n    # Save user in the Firestore db\n    await frantic.add(user)\n\n    # user\'s id gets automatically populated and can be used to retrieve the user\n    retrieved = await frantic.get(User, user.id)\n    assert retrieved.name == user.name\n\n    # list all users\n    users = await frantic.list(User)\n\n    # delete user\n    await frantic.delete(user)\n    # or\n    await frantic.delete(User, user.id)\n\nif __name__ == "__main__":\n    asyncio.run(main)\n```\n',
    'author': 'Tomas Votava',
    'author_email': 'info@tomasvotava.eu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tomasvotava/frantic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
